#!/usr/bin/env python
#
# (C) Copyright 2017- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation nor
# does it submit to any jurisdiction.
#

import io
import os
import re
import sys

import setuptools

binary = os.getenv("DEPEND_ON_ECCODESLIB", "") == "1"


def read(path):
    file_path = os.path.join(os.path.dirname(__file__), *path.split("/"))
    return io.open(file_path, encoding="utf-8").read()


# single-sourcing the package version using method 1 of:
#   https://packaging.python.org/guides/single-sourcing-package-version/
def parse_version_from(path):
    version_pattern = (
        r"^__version__ = [\"\'](.*)[\"\']"  # More permissive regex pattern
    )
    version_file = read(path)
    version_match = re.search(version_pattern, version_file, re.M)
    if version_match is None or len(version_match.groups()) > 1:
        raise ValueError("couldn't parse version")
    return version_match.group(1)


# for the binary wheel
libdir = os.path.realpath("install/lib")
incdir = os.path.realpath("install/include")
libs = ["eccodes"]

if "--binary-wheel" in sys.argv:
    sys.argv.remove("--binary-wheel")

    # https://setuptools.pypa.io/en/latest/userguide/ext_modules.html
    ext_modules = [
        setuptools.Extension(
            "eccodes._eccodes",
            sources=["eccodes/_eccodes.cc"],
            language="c++",
            libraries=libs,
            library_dirs=[libdir],
            include_dirs=[incdir],
            extra_link_args=["-Wl,-rpath," + libdir],
        )
    ]

    def shared(directory):
        result = []
        for path, dirs, files in os.walk(directory):
            for f in files:
                result.append(os.path.join(path, f))
        return result

    # Paths must be relative to package directory...
    shared_files = ["versions.txt"]
    shared_files += [x[len("eccodes/") :] for x in shared("eccodes/copying")]

    if os.name == "nt":
        for n in os.listdir("eccodes"):
            if n.endswith(".dll"):
                shared_files.append(n)

else:
    ext_modules = []
    shared_files = []


install_requires = ["numpy"]
if sys.version_info < (3, 7):
    install_requires = ["numpy<1.20"]
elif sys.version_info < (3, 8):
    install_requires = ["numpy<1.22"]
elif sys.version_info < (3, 9):
    install_requires = ["numpy<1.25"]

install_requires += [
    "attrs",
    "cffi",
    "findlibs",
]

# only add eccodeslib as dependency if building a distribution that should depend on it
if binary:
    install_requires += [
        "eccodeslib;platform_system!='Windows'",
    ]

setuptools.setup(
    name="eccodes",
    version=parse_version_from("gribapi/bindings.py"),
    description="Python interface to the ecCodes GRIB and BUFR decoder/encoder",
    long_description=read("README.rst") + read("CHANGELOG.rst"),
    author="European Centre for Medium-Range Weather Forecasts (ECMWF)",
    author_email="software.support@ecmwf.int",
    license="Apache License Version 2.0",
    url="https://github.com/ecmwf/eccodes-python",
    packages=setuptools.find_packages(),
    include_package_data=True,
    package_data={"": shared_files},
    install_requires=install_requires,
    tests_require=[
        "pytest",
        "pytest-cov",
        "pytest-flakes",
    ],
    test_suite="tests",
    zip_safe=True,
    keywords="ecCodes GRIB BUFR",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Operating System :: OS Independent",
    ],
    ext_modules=ext_modules,
)



# Copyright 2022- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

# flake8: noqa: F403
# flake8: noqa: F405

import datetime as dt
import sys

import numpy as np
import pytest as pt

from eccodes import *
from eccodes.highlevel._bufr import *

if sys.platform.startswith("win32"):
    pt.skip("Not applicable on Windows", allow_module_level=True)


@pt.fixture(autouse=True)
def fixture(monkeypatch):
    import os

    this_dir = os.path.dirname(__file__)
    monkeypatch.chdir(this_dir)  # always run tests from the tests directory


def unpacked(bufr):
    defined = codes_is_defined(bufr._handle, "year")
    return bool(defined)


def compare_items(m1, m2, keys=None):
    if keys:
        keys = list(keys)
        i1 = [(k, m1[k]) for k in keys]
        i2 = [(k, m2[k]) for k in keys]
    else:
        i1 = list(m1.items())
        i2 = list(m2.items())
    for (k1, v1), (k2, v2) in zip(i1, i2):
        assert k1 == k2
        if isinstance(v1, np.ndarray):
            if v1.dtype.type == np.str_:
                equal = np.char.compare_chararrays(v1, v2, "==", rstrip=True)
            else:
                equal = v1.data == v2.data
            assert k1 and np.all(equal)
        else:
            assert k1 and v1 == v2


def test_message_uncompressed():
    bufr = BUFRMessage(open("./sample-data/synop_multi_subset.bufr"))
    assert bufr["edition"] == 4
    assert np.all(
        bufr["unexpandedDescriptors"] == [307079, 4025, 11042]
    )  # is always array
    assert "dataCategory" in bufr.header
    assert bufr.get_count("dataCategory") == 1
    bufr["dataCategory"] = 11
    assert not unpacked(bufr)
    assert "rubbish_key" not in bufr  # triggers unpacking
    assert unpacked(bufr)
    # Accessing undefined elements raises an exception
    with pt.raises(KeyValueNotFoundError):
        bufr["undefined"]
    with pt.raises(KeyValueNotFoundError):
        bufr["undefined"] = 0
    assert "blockNumber" in bufr
    assert bufr.get_count("blockNumber") == 12
    assert len(bufr.data) == 12  # number of subsets
    assert len(bufr.data[0]) == 19  # number of data tree nodes per subset
    assert np.all(
        bufr["stationNumber"].data
        == np.array([27, 84, 270, 272, 308, 371, 381, 382, 387, 413, 464, 485])
    )
    assert bufr["#1#stationNumber"] == 27
    assert bufr.data[0]["stationNumber"] == 27  # of the 1st subset
    bufr["stationNumber"][0] = missing_of(int)
    assert bufr["stationNumber"].data[0] == missing_of(int)
    bufr["stationNumber"] = missing_of(int)
    assert np.all(bufr["stationNumber"].data == missing_of(int))
    # Keys can have element attribute as a suffix (à la ecCodes)
    assert bufr["stationNumber->units"] == "Numeric"
    # Attempting to modify element attributes raises an exception
    with pt.raises(ReadOnlyError):
        bufr["stationNumber->units"] = "Turmeric"


def test_message_compressed():
    bufr = BUFRMessage(open("./sample-data/sral_sentinel_3a.bufr"))
    assert (
        len(bufr.data) == 1
    )  # this message doesn't have any replications, so the structure is flat
    year = bufr["year"]
    assert year.shape == (22, 5)
    assert np.all(year[0].data == 2017)
    assert np.all(year[-1].data == missing_of(int))
    # Keys can have rank prefix (à la ecCodes)
    bufr["#1#year"] = 2018
    assert np.all(bufr["#1#year"].data == 2018)
    # Packing the message by setting the 'pack' key is disallowed
    with pt.raises(KeyValueNotFoundError):
        bufr["pack"] = 1
    # The same goes for unpacking
    with pt.raises(KeyValueNotFoundError):
        bufr["unpack"] = 1


def test_message_sample_defaults():
    bufr = BUFRMessage("BUFR3")
    assert bufr["edition"] == 3
    assert bufr["masterTableNumber"] == 0
    assert bufr["masterTablesVersionNumber"] == 24
    assert bufr["localTablesVersionNumber"] == 0
    assert bufr["bufrHeaderCentre"] == 98
    assert bufr["bufrHeaderSubCentre"] == 0
    assert bufr["numberOfSubsets"] == 1
    assert not bufr["compressedData"]
    assert not bufr["section2Present"]
    assert np.all(bufr["unexpandedDescriptors"] == 307080)
    bufr = BUFRMessage("BUFR3_local")
    assert bufr["edition"] == 3
    assert bufr["masterTableNumber"] == 0
    assert bufr["masterTablesVersionNumber"] == 24
    assert bufr["localTablesVersionNumber"] == 0
    assert bufr["bufrHeaderCentre"] == 98
    assert bufr["bufrHeaderSubCentre"] == 0
    assert bufr["numberOfSubsets"] == 1
    assert not bufr["compressedData"]
    assert bufr["section2Present"]
    assert np.all(bufr["unexpandedDescriptors"] == 307080)
    bufr = BUFRMessage("BUFR4_local")
    assert bufr["edition"] == 4
    assert bufr["masterTableNumber"] == 0
    assert bufr["masterTablesVersionNumber"] == 24
    assert bufr["localTablesVersionNumber"] == 0
    assert bufr["bufrHeaderCentre"] == 98
    assert bufr["bufrHeaderSubCentre"] == 0
    assert bufr["numberOfSubsets"] == 1
    assert not bufr["compressedData"]
    assert np.all(bufr["unexpandedDescriptors"] == 307080)
    bufr = BUFRMessage("BUFR3_local_satellite")
    assert bufr["edition"] == 3
    assert bufr["masterTableNumber"] == 0
    assert bufr["masterTablesVersionNumber"] == 13
    assert bufr["localTablesVersionNumber"] == 1
    assert bufr["bufrHeaderCentre"] == 98
    assert bufr["bufrHeaderSubCentre"] == 0
    assert bufr["numberOfSubsets"] == 128
    assert bufr["compressedData"]
    assert np.all(bufr["unexpandedDescriptors"] == 310014)
    bufr = BUFRMessage("BUFR4_local_satellite")
    assert bufr["edition"] == 4
    assert bufr["masterTableNumber"] == 0
    assert bufr["masterTablesVersionNumber"] == 13
    assert bufr["localTablesVersionNumber"] == 1
    assert bufr["bufrHeaderCentre"] == 98
    assert bufr["bufrHeaderSubCentre"] == 0
    assert bufr["numberOfSubsets"] == 128
    assert bufr["compressedData"]
    assert np.all(bufr["unexpandedDescriptors"] == 310008)


def test_message_new_from_sample():
    bufr = BUFRMessage("BUFR3_local_satellite")
    # If the tables version is not set properly, eccodes fails with
    # "HashArrayNoMatchError: Hash array no match" when setting
    # unexpandedDescriptors.
    bufr["masterTablesVersionNumber"] = 18
    bufr["localTablesVersionNumber"] = 0
    bufr["numberOfSubsets"] = 2
    bufr["compressedData"]
    # Input replication factors are optional; will default to 0 if not set.
    bufr["inputDelayedDescriptorReplicationFactor"] = [0, 0]
    bufr["inputShortDelayedDescriptorReplicationFactor"] = [0, 0, 0, 0, 0, 0]
    bufr["unexpandedDescriptors"] = [311011]

    bufr["year"] = 2022

    # Template-related items can't be changed once the header is "baked" in
    for key in ("numberOfSubsets", "compressedData", "unexpandedDescriptors"):
        with pt.raises(ReadOnlyError):
            bufr[key] = 1
    # Test that common Section 2 keys are set properly
    bufr.pack()
    assert bufr["localNumberOfObservations"] == bufr["numberOfSubsets"]


def test_message_new_from_bytes():
    old = BUFRMessage("BUFR3_local")
    bytes = old.get_buffer()
    new = BUFRMessage(bytes)
    compare_items(old.header, new.header, keys=old.header.keys(skip="read_only"))
    compare_items(old.data, new.data)


def test_message_context_manager():
    bufr = BUFRMessage("BUFR3_local")
    assert bufr._handle > 0
    with bufr:
        pass
    assert bufr._handle == 0


def test_message_copy():
    # Copy is the same as the original (note that copying doesn't require unpacking)
    old = BUFRMessage(open("./sample-data/synop.bufr"))
    assert unpacked(old) is False
    new = old.copy()
    assert unpacked(old) is False
    assert old["md5Data"] == new["md5Data"]
    assert old._handle != new._handle
    compare_items(old, new)

    # FIXME: The following code doesnt't work because eccodes apparently doesn't
    # update typicalDate (which is computed key) when we modify typicalYearOfCentury .

    # After chainging one of the header section keys
    # old['typicalYearOfCentury'] = 25
    # new = old.copy()
    # assert old['md5Data'] == new['md5Data']
    # compare_items(old, new)

    # After changing one of the data section keys
    old["blockNumber"] = 99
    new = old.copy()
    assert old["md5Data"] == new["md5Data"]
    compare_items(old, new)


def test_message_copy_subsets():
    old = BUFRMessage(open("./sample-data/ahi-himawari-8.bufr"))
    old_count = old["numberOfSubsets"]
    old["second"] = range(0, old_count)
    # Copy adjacent subsets where mask is True (uses extractSubsetInterval)
    mask = np.repeat(False, old_count)
    mask[0:3] = True
    with old.copy(subsets=mask) as new:
        assert new["numberOfSubsets"] == 3
        assert np.all(new["second"] == [0, 1, 2])
    # # Copy adjacent subsets where mask is True (uses extractSubsetList)
    mask = np.repeat(False, old_count)
    mask[[0, 2, 4]] = True
    with old.copy(subsets=mask) as new:
        assert new["numberOfSubsets"] == 3
        assert np.all(new["second"] == [0, 2, 4])
    # Copy a range of adjacent subsets (uses extractSubsetInterval)
    with old.copy(subsets=range(3)) as new:
        assert new["numberOfSubsets"] == 3
        assert np.all(new["second"] == [0, 1, 2])
    # Copy a range of non-adjacent subsets (uses extractSubsetList)
    with old.copy(subsets=range(0, 5, 2)) as new:
        assert new["numberOfSubsets"] == 3
        assert np.all(new["second"] == [0, 2, 4])
    # Copy a list of adjacent subsets (uses extractSubsetInterval)
    with old.copy(subsets=[0, 1, 2]) as new:
        assert new["numberOfSubsets"] == 3
        assert np.all(new["second"] == [0, 1, 2])
    # Copy a list of non-adjacent subsets (uses extractSubsetList)
    with old.copy(subsets=[0, 2, 4]) as new:
        assert new["numberOfSubsets"] == 3
        assert np.all(new["second"] == [0, 2, 4])
    # Copy a full slice of all subsets
    # (uses extractSubsetInterval, or no extraction if there were no previous ones)
    with old.copy(subsets=slice(None, None, None)) as new:  # [1]
        assert new["numberOfSubsets"] == old_count
        assert np.all(new["second"] == range(0, old_count))
    # Copy every 4th subset (uses extractSubsetInterval)
    with old.copy(subsets=slice(None, None, 4)) as new:
        assert new["numberOfSubsets"] == divmod(old_count, 4)[0]
        assert np.all(new["second"] == range(0, old_count, 4))
    # Copy every 4th subset within a range (uses extractSubsetList)
    with old.copy(subsets=slice(16, 32, 4)) as new:
        assert new["numberOfSubsets"] == divmod(32 - 16, 4)[0]
        assert np.all(new["second"] == old["second"][16:32:4])
    # Copy datetime slice (uses extractSubsetInterval)
    old_datetime = old.get_datetime()
    start = old_datetime[0].item()
    end = start + dt.timedelta(seconds=2)
    with old.copy(subsets=slice(None, end)) as new:
        assert new["numberOfSubsets"] == 3
        new_datetime = new.get_datetime()
        assert np.all(new_datetime == old_datetime[0:3])

    # [1] Note that this prints a lot of 'ECCODES ERROR' messages due to ECC-2025,
    #     but otherwise the code works OK because we work around the issue internally.


def test_message_copy_to():
    old = BUFRMessage(open("./sample-data/aura-omi-ak.bufr"))
    new = BUFRMessage("BUFR3_local_satellite")
    old.copy_to(new)
    compare_items(old.header, new.header, keys=old.header.keys(skip="read_only"))
    compare_items(old.data, new.data)


def test_message_eq():
    bufr1 = BUFRMessage(open("./sample-data/synop.bufr"))
    bufr2 = BUFRMessage(open("./sample-data/synop.bufr"))
    assert bufr1 == bufr2
    bufr1["typicalDay"] = 1
    bufr2["typicalDay"] = 2
    assert bufr1.header != bufr2.header
    assert bufr1.data == bufr2.data


def test_message_update():
    bufr = BUFRMessage("BUFR3")
    bufr.update(year=2000)
    bufr.update({"month": 2})
    bufr.update([("day", 4)])
    bufr.update(dict(hour=8, minute=16).items())
    assert bufr["year"] == 2000
    assert bufr["month"] == 2
    assert bufr["day"] == 4
    assert bufr["hour"] == 8
    assert bufr["minute"] == 16


def test_message_get_datetime_compressed():
    bufr = BUFRMessage("BUFR4_local_satellite")
    bufr.update(numberOfSubsets=2)
    datetime = bufr.get_datetime()
    assert isinstance(datetime, np.ndarray)
    assert datetime.size == 2
    assert datetime.all() is np.ma.masked
    bufr.update(year=2000, month=1, day=2, hour=3, minute=4)
    bufr.update(second=[5, CODES_MISSING_LONG])
    datetime = bufr.get_datetime()
    assert isinstance(datetime, np.ndarray)
    assert datetime.size == 2
    assert np.ma.all(
        datetime == np.array(["2000-01-02T03:04:05", "NaT"], dtype="M8[s]")
    )


def test_message_get_datetime_uncompressed():
    bufr = BUFRMessage("BUFR3")
    datetime = bufr.get_datetime()
    assert datetime is np.ma.masked
    bufr.update(year=2000, month=1, day=2, hour=3, minute=4)
    datetime = bufr.get_datetime()
    assert not isinstance(datetime, np.ndarray)
    assert datetime == np.datetime64("2000-01-02T03:04:00")
    datetime = bufr.get_datetime(prefix="typical")
    assert datetime == np.datetime64("2012-10-31T00:02:00")


def test_message_set_datetime_compressed():
    bufr = BUFRMessage("BUFR4_local_satellite")
    # bufr.update(numberOfSubsets=3, unexpandedDescriptors=310008)
    bufr.update(numberOfSubsets=3)
    # Built-in datetime object
    datetime = np.datetime64("2000-01-02T03:04:05", "s").item()
    bufr.set_datetime(datetime)
    assert np.all(bufr["year"] == 2000)
    assert np.all(bufr["month"] == 1)
    assert np.all(bufr["day"] == 2)
    assert np.all(bufr["hour"] == 3)
    assert np.all(bufr["minute"] == 4)
    assert np.all(bufr["second"] == 5)
    # Numpy datetime64 array
    values = ["2000-01-02T03:04:05", "2000-01-02T06:07:08", "NaT"]
    array = np.ma.array(values, dtype="M8[s]", mask=[False, False, True])
    bufr.set_datetime(array)
    assert np.all(bufr["year"] == 2000)
    assert np.all(bufr["month"] == 1)
    assert np.all(bufr["day"] == 2)
    assert np.all(bufr["hour"] == np.array([3, 6, -1]))
    assert np.all(bufr["minute"] == np.array([4, 7, -1]))
    assert np.all(bufr["second"] == np.array([5, 8, -1]))
    # Unless explicitly set by the user, all datetime-related keys from the header
    # section (i.e., typicalYear, typicalMonth, etc.) get updated automatically,
    # whilst packing, based on values from the data section.
    bufr["typicalYear"]
    bufr["localYear"]
    bufr.pack()
    assert bufr["typicalYear"] == 2000
    assert bufr["typicalMonth"] == 1
    assert bufr["typicalDay"] == 2
    assert bufr["typicalHour"] == 3
    assert bufr["typicalMinute"] == 4
    assert bufr["typicalSecond"] == 5
    assert bufr["localYear"] == 2000
    assert bufr["localMonth"] == 1
    assert bufr["localDay"] == 2
    assert bufr["localHour"] == 3
    assert bufr["localMinute"] == 4
    assert bufr["localSecond"] == 5
    # But respect values of section 1 keys if explicitly set by the user.
    bufr.update(typicalMinute=0, typicalSecond=0)
    bufr.update(localMinute=0, localSecond=0)
    bufr.pack()
    assert bufr["typicalMinute"] == 0
    assert bufr["typicalSecond"] == 0
    assert bufr["localMinute"] == 0
    assert bufr["localSecond"] == 0


def test_message_set_datetime_compressed_irregular():
    bufr = BUFRMessage("BUFR3_local_satellite")
    bufr.update(numberOfSubsets=3)
    assert bufr.get_count("year") == 2
    assert bufr.get_count("month") == 2
    assert bufr.get_count("day") == 2
    assert bufr.get_count("hour") == 10
    assert bufr.get_count("minute") == 9
    assert bufr.get_count("second") == 9
    values = ["2000-01-02T03:04:05", "2000-01-02T06:07:08", "NaT"]
    array = np.ma.array(values, dtype="M8[s]", mask=[False, False, True])
    bufr.set_datetime(array, rank=1)
    assert np.all(bufr["#1#year"] == 2000)
    assert np.all(bufr["#1#month"] == 1)
    assert np.all(bufr["#1#day"] == 2)
    assert np.all(bufr["#1#hour"] == np.array([3, 6, -1]))
    assert np.all(bufr["#1#minute"] == np.array([4, 7, -1]))
    assert np.all(bufr["#1#second"] == np.array([5, 8, -1]))
    assert np.all(bufr["year"][1:]) is np.ma.masked
    assert np.all(bufr["month"][1:]) is np.ma.masked
    assert np.all(bufr["day"][1:]) is np.ma.masked
    assert np.all(bufr["hour"][1:]) is np.ma.masked
    assert np.all(bufr["minute"][1:]) is np.ma.masked
    assert np.all(bufr["second"][1:]) is np.ma.masked
    # Section 1 keys updated whilst packing
    bufr["typicalYearOfCentury"]
    bufr.pack()
    assert bufr.get_datetime(prefix="typical") == np.datetime64(values[0])
    assert bufr.get_datetime(prefix="local") == np.datetime64(values[0])


def test_message_set_datetime_uncompressed():
    bufr = BUFRMessage("BUFR3_local")
    datetime = np.datetime64("2000-01-02T03:04:05", "s").item()
    bufr.set_datetime(datetime)
    assert bufr["year"] == 2000
    assert bufr["month"] == 1
    assert bufr["day"] == 2
    assert bufr["hour"] == 3
    assert bufr["minute"] == 4
    # Section 1 keys updated whilst packing
    bufr["typicalYearOfCentury"]
    bufr.pack()
    assert bufr["typicalYearOfCentury"] == 0
    assert bufr["typicalMonth"] == 1
    assert bufr["typicalDay"] == 2
    assert bufr["typicalHour"] == 3
    assert bufr["typicalMinute"] == 4
    assert bufr["typicalSecond"] == 0



#
# (C) Copyright 2017- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation nor
# does it submit to any jurisdiction.
#

# Authors:
#   Alessandro Amici - B-Open - https://bopen.eu
#   Shahram Najm - ECMWF - https://www.ecmwf.int
#

from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import os
import pkgutil
import sys

import cffi

__version__ = "2.45.0"

LOG = logging.getLogger(__name__)

_MAP = {
    "grib_api": "eccodes",
    "gribapi": "eccodes",
}

EXTENSIONS = {
    "darwin": ".dylib",
    "win32": ".dll",
}

# convenient way to trace the search for the library
if int(os.environ.get("ECCODES_PYTHON_TRACE_LIB_SEARCH", "0")):
    LOG.setLevel(logging.DEBUG)
    LOG.addHandler(logging.StreamHandler())


def _lookup(name):
    return _MAP.get(name, name)


def find_binary_libs(name):
    name = _lookup(name)
    env_var = "ECCODES_PYTHON_USE_FINDLIBS"
    if int(os.environ.get(env_var, "0")):
        LOG.debug(f"{name} lib search: {env_var} set, so using findlibs")

    elif os.name == "nt":
        LOG.debug(f"{name} lib search: trying to find binary wheel")
        here = os.path.dirname(__file__)
        # eccodes libs are actually in eccodes dir, not gribapi dir
        here = os.path.abspath(os.path.join(here, os.path.pardir, "eccodes"))
        extension = EXTENSIONS.get(sys.platform, ".so")

        for libdir in [here + ".libs", os.path.join(here, ".dylibs"), here]:
            LOG.debug(f"{name} lib search: looking in {libdir}")
            if not name.startswith("lib"):
                libnames = ["lib" + name, name]
            else:
                libnames = [name, name[3:]]

            if os.path.exists(libdir):
                for file in os.listdir(libdir):
                    if file.endswith(extension):
                        for libname in libnames:
                            if libname == file.split("-")[0].split(".")[0]:
                                foundlib = os.path.join(libdir, file)
                                LOG.debug(
                                    f"{name} lib search: returning wheel library from {foundlib}"
                                )
                                # force linking with the C++ 'glue' library
                                try:
                                    from eccodes._eccodes import versions as _versions
                                except ImportError as e:
                                    LOG.warn(str(e))
                                    raise
                                LOG.debug(
                                    f"{name} lib search: versions: %s", _versions()
                                )
                                return foundlib

        LOG.debug(
            f"{name} lib search: did not find library from wheel; try to find as separate lib"
        )

    # if did not find the binary wheel, or the env var is set, fall back to findlibs
    import findlibs

    foundlib = findlibs.find(name)
    LOG.debug(f"{name} lib search: findlibs returned {foundlib}")
    return foundlib


library_path = find_binary_libs("eccodes")

if library_path is None:
    raise RuntimeError("Cannot find the ecCodes library")

# default encoding for ecCodes strings
ENC = "ascii"

ffi = cffi.FFI()
CDEF = pkgutil.get_data(__name__, "grib_api.h")
CDEF += pkgutil.get_data(__name__, "eccodes.h")
ffi.cdef(CDEF.decode("utf-8").replace("\r", "\n"))


lib = ffi.dlopen(library_path)



# Copyright 2022- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from functools import cached_property

from .coder import Coder
from .common import *
from .helpers import ensure_masked_array, flatten, missing_of
from .tree import (
    AssociationNode,
    LeafNode,
    Node,
    ReplicationNode,
    WrapperNode,
    build_tree,
)
from .view import View

# flake8: noqa: F405
#   ruff: noqa: F403


class Data(View):
    _coder: Coder
    _subset_count: int = 0
    _compressed: bool = False
    _entries: Dict[str, DataEntry]
    _associations: Dict[Tuple[str, int], Association]

    def __init__(self, coder: Coder) -> None:
        self._coder = coder
        self._entries = {}

    def __contains__(self, key: str) -> bool:
        return self._top_view.__contains__(key)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Data):
            eq = self._top_view.__eq__(other._top_view)
        else:
            eq = False
        return eq

    def __getitem__(self, subscript: Union[int, str]) -> Union["DataBlock", ValueLike]:
        return self._top_view.__getitem__(subscript)

    def __iter__(self) -> Iterator["DataBlock"]:
        yield from self._top_view.__iter__()

    def __len__(self) -> int:
        return self._top_view.__len__()

    def __setitem__(self, key: str, value: ValueLike) -> None:
        return self._top_view.__setitem__(key, value)

    def __str__(self) -> str:
        return self._top_view.__str__()

    def as_dict(self, ranked=False, depth=0, **kwds) -> Dict:
        """Returns dict-like representation of the data section items."""
        return self._top_view.as_dict(ranked, depth, **kwds)

    def copy_to(self, other: "Data") -> None:
        """Copies common keys/values in this data section to `other`."""
        self._coder.unpack()
        self._commit()
        codes_bufr_copy_data(self._coder._handle, other._coder._handle)
        # Workaround for ECC-2022
        self._top_view  # [1]
        for a in self._associations.values():
            if a.element_rank > 1:
                for g in a.entries.values():
                    other[g.name][:] = self[g.name][:]

        # [1] This ensures that we have built all bitmap associations.

    def get_count(self, key: str) -> int:
        """Returns the number of ranked items designated by the given key."""
        return self._top_view.get_count(key)

    def get_shape(self, key: str) -> Tuple[int, ...]:
        return self._top_view.get_shape(key)

    def get_size(self, key: str) -> int:
        return self._top_view.get_size(key)

    def is_missing(self, key: str) -> bool:
        return self._top_view.is_missing(key)

    def items(self, ranked=False, **kwds) -> Iterator[Tuple[str, ValueLike]]:
        yield from self._top_view.items(ranked, **kwds)

    def keys(self, ranked=False, **kwds) -> Iterator[str]:
        yield from self._top_view.keys(ranked, **kwds)

    def set(self, key: str, value: ValueLike) -> None:
        self._top_view.set(key, value)

    def set_missing(self, key: str) -> None:
        self._top_view.set_missing(key)

    def _commit(self) -> None:
        for entry in self._entries.values():
            if entry.array is None or entry.flags & (READ_ONLY | COMPUTED):
                continue
            else:
                self._coder.commit(entry)
                entry.array = None

    @cached_property
    def _top_view(self):
        self._subset_count = cast(int, self._coder.get("numberOfSubsets"))
        self._compressed = cast(bool, self._coder.get("compressedData"))
        root, self._entries, self._associations = build_tree(self._coder)  # [1]
        top_level = ()
        top_view = DataBlock(self, root, top_level)
        for entry in self._entries.values():
            try:
                slice = top_view._slices[entry.name]
            except KeyError:
                continue
            if entry.association:
                entry.shape = entry.primary.shape
            elif entry.flags & SCALAR:
                entry.shape = (slice.stop, 1)
                if entry.flags & FACTOR:
                    assert entry.array is not None  # [2]
            elif entry.flags & BITMAP:
                entry.shape = (slice.stop, None)  # [3]
            else:
                if self._compressed:
                    entry.shape = (slice.stop, self._subset_count)
                else:
                    entry.shape = (slice.stop,)
        self._coder._baked_template = True
        self._coder._subset_count = self._subset_count
        self._coder._compressed = self._compressed
        return top_view

        # [1] Note that it's the call to build_tree() which triggers unpacking!
        #     Because _top_view() is a cached property (and thus initialized lazily)
        #     the unpacking will be defered until when actually needed.
        #
        # [2] Note that arrays of delayed replication factors are constructed in build_tree().
        #
        # [3] At the moment the size of bitmap arrays is determined in Coder.checkout().
        #     Is there a way to infer it earlier? TODO


class DataBlock(View):
    def __init__(self, data: Data, node, index: MultiIndex) -> None:
        self._data: Data = data
        self._node = node
        self._index: MultiIndex = index
        self._slices: Dict[str, slice] = resolve_slices(node, index)
        self._sub_blocks: Dict[int, "DataBlock"] = {}

    def __contains__(self, subscript: str) -> bool:
        key = Key.from_string(subscript)
        try:
            entry = self._get_entry(key)
        except NotFoundError:
            contains = False
        else:
            if key.rank:
                slice = self._slices[entry.name]
                contains = key.rank <= slice.stop
            else:
                contains = True
        return contains

    def __eq__(self, other: object) -> bool:
        if isinstance(other, DataBlock):
            eq = False
            for (k1, v1), (k2, v2) in zip(self.items(), other.items()):
                if k1 != k2:
                    break
                if isinstance(v1, np.ndarray):
                    if not isinstance(v2, np.ndarray):
                        break
                    if v1.dtype.type == np.str_:
                        if not v2.dtype.type == np.str_:
                            break
                        if not np.char.compare_chararrays(v1, v2, "==", rstrip=True):
                            break
                    else:
                        if not np.all(v1.data == v2.data):
                            break
                else:
                    if v1 != v2:
                        break
            else:
                eq = True
        else:
            eq = False
        return eq

    def __len__(self) -> int:
        node = self._node
        if isinstance(node, AssociationNode):
            # Has no children. Should we add an empty list? TODO
            return 0
        elif node.children and isinstance(node.children[0], ReplicationNode):
            assert len(node.children) == 1
            replication = node.children[0]
            return replication.factors[self._index]
        else:
            return len(node.children)

    def __iter__(self) -> Iterator["DataBlock"]:
        node = self._node
        if node.children and isinstance(node.children[0], ReplicationNode):
            assert len(node.children) == 1
            replication = node.children[0]
            for factor in range(replication.factors[self._index]):
                index = self._index + (factor,)
                if len(replication.children) == 1 and isinstance(
                    replication.children[0], LeafNode
                ):
                    block = DataBlock(self._data, replication.children[0], index)
                else:
                    block = DataBlock(self._data, replication, index)
                yield block
        else:
            for child in node.children:
                yield DataBlock(self._data, child, self._index)

    def __getitem__(self, subscript: Union[int, str]) -> Union["DataBlock", ValueLike]:
        if isinstance(subscript, str):
            key = Key.from_string(subscript)
            entry = self._get_entry(key)
            sub = self._get_subscript(entry, key)
            if key.attribute:
                if entry.uniform_element:
                    value = getattr(entry.uniform_element, key.attribute)
                else:
                    if not entry.elements:
                        resolve_elements(
                            self._data._entries, self._data._top_view._node
                        )
                        assert entry.elements
                    if isinstance(sub, slice):
                        value = [getattr(e, key.attribute) for e in entry.elements[sub]]
                    else:
                        value = getattr(entry.elements[sub], key.attribute)
            else:
                array = self._get_array(entry)
                value = array[sub]
            return value
        elif type(ordinal := subscript) is int:
            node = self._node
            if node.children and isinstance(node.children[0], ReplicationNode):
                assert len(node.children) == 1
                replication = node.children[0]
                factor = replication.factors[self._index]
                ordinal = factor + ordinal if ordinal < 0 else ordinal
                if ordinal < 0 or ordinal >= factor:
                    raise IndexError(f"Subscript out of range: {subscript}")
                try:
                    block = self._sub_blocks[ordinal]
                except KeyError:
                    index = self._index + (ordinal,)
                    if len(replication.children) == 1 and isinstance(
                        replication.children[0], LeafNode
                    ):
                        block = DataBlock(self._data, replication.children[0], index)
                    else:
                        block = DataBlock(self._data, replication, index)
                    self._sub_blocks[ordinal] = block
            else:
                ordinal = len(node.children) + ordinal if ordinal < 0 else ordinal
                if ordinal < 0 or ordinal >= len(node.children):
                    raise IndexError(f"Subscript out of range: {subscript}")
                try:
                    block = self._sub_blocks[ordinal]
                except KeyError:
                    block = DataBlock(self._data, node.children[ordinal], self._index)
                    self._sub_blocks[ordinal] = block
            return block
        else:
            raise TypeError(
                "subscript must be either 'int' or 'str'; got %s" % type(subscript)
            )

    def __setitem__(self, key: str, value: ValueLike) -> None:
        key = Key.from_string(key)
        value = ensure_masked_array(value)
        entry = self._get_entry(key)
        if entry.flags & READ_ONLY or key.attribute:
            raise ReadOnlyError(f"{key.string}")
        array = self._get_array(entry)
        subscript = self._get_subscript(entry, key)
        if isinstance(subscript, slice):
            array[subscript] = value
        else:
            array[subscript : subscript + 1] = value

    def __str__(self) -> str:
        import pprint

        d = self.as_dict()
        return pprint.pformat(d, sort_dicts=False)

    def as_dict(self, ranked=False, depth=0, **kwds) -> Dict:
        """Returns dict-like representation of the DataBlock items."""
        if self._node.depth < depth and len(self) > 0:
            d = {}
            for i, block in enumerate(self):
                # block = cast('DataBlock', self[i])
                d[i] = block.as_dict(ranked, depth, **kwds)
        else:
            counts_only = kwds.pop("counts_only", False)
            shapes_only = kwds.pop("shapes_only", False)
            if counts_only:
                d = {k: self.get_count(k) for k in self.keys(ranked, **kwds)}
            elif shapes_only:
                d = {k: self.get_shape(k) for k in self.keys(ranked, **kwds)}
            else:
                d = dict(self.items(ranked, **kwds))
        return d

    def get_count(self, key: str) -> int:
        """Returns the number of elements designated by the given key."""
        key = Key.from_string(key)
        try:
            slice = self._slices[key.name]
        except KeyError:
            raise NotFoundError(key.string)
        else:
            rank_count = slice.stop - slice.start
            if key.rank:
                if key.rank <= rank_count:
                    count = 1
                else:
                    raise NotFoundError(key.string)
            else:
                count = rank_count
        return count

    def get_shape(self, key: str) -> Tuple[int, ...]:
        key = Key.from_string(key)
        entry = self._get_entry(key)
        subscript = self._get_subscript(entry, key)
        if isinstance(subscript, slice):
            shape = (subscript.stop - subscript.start,)
        else:
            shape = ()
        if key.attribute:
            if entry.uniform_element:
                shape = ()
        elif self._data._compressed and not entry.flags & Flags.SCALAR:
            shape += (self._data._subset_count,)
        return shape

    def get_size(self, key: str) -> int:
        shape = self.get_shape(key)
        size = 1
        for dim in shape:
            size *= dim
        return size

    def is_missing(self, key: str) -> bool:
        key = Key.from_string(key)
        if key.attribute:
            is_missing = False
        else:
            entry = self._get_entry(key)
            slice = self._get_slice(entry, key)
            if entry.array is None:
                is_missing = self._data._coder.is_missing(entry, slice)
            else:
                array = self._get_array(entry)
                array_view = array[slice]
                is_missing = np.all(array_view.mask)
        return is_missing

    def set(self, key: str, value: ValueLike) -> None:
        self.__setitem__(self, key, value)

    def set_missing(self, key: str) -> None:
        key = Key.from_string(key)
        if key.attribute:
            raise ValueCannotBeMissingError(
                f"Attribute '{key.string}' can't have missing value"
            )
        else:
            entry = self._get_entry(key)
            slice = self._get_slice(entry, key)
            if entry.array is None:
                self._data._coder.set_missing(entry, slice)
            else:
                array = self._get_array(entry)
                array_view = array[slice]
                array_view.mask = True

    def items(self, ranked=False, **kwds) -> Iterator[Tuple[str, ValueLike]]:
        for key in self.keys(ranked, **kwds):
            yield key, cast(ValueLike, self[key])

    def keys(self, ranked=False, **kwds) -> Iterator[str]:
        only_flags = ensure_flags(kwds.get("only", Flags.CODED))
        skip_flags = ensure_flags(kwds.get("skip", Flags(0)))
        if ranked:
            ranks = Counter()
            for key in iterate_keys(self._node, self._index):
                ranks[key.name] += 1
                rank = ranks[key.name]
                yield f"#{rank}#{key.name}"
        else:
            for name, slice in self._slices.items():
                if slice.stop <= slice.start:
                    continue  # Ignore keys inside empty replications
                try:
                    entry = self._data._entries[name]
                except KeyError:
                    continue  # Ignore unset associated keys
                if not entry.flags & only_flags:
                    continue
                if entry.flags & skip_flags:
                    continue
                yield name

    def _get_entry(self, key: Key) -> DataEntry:
        try:
            entry = self._data._entries[key.name]
        except KeyError:
            raise NotFoundError(key.string)
        return entry

    def _get_array(self, entry):
        if entry.array is None:
            array = self._data._coder.checkout(entry)
            if not self._data._coder._compressed:  # TODO: correct shape in Coder
                array = array.ravel()
            if array.dtype.type == np.str_ and np.__version__ < "1.24":
                array = np.ma.masked_where(array == "", array, copy=False)  # [1]
                array.fill_value = ""
            else:
                array = np.ma.masked_equal(array, missing_of(array.dtype), copy=False)
            entry.array = array
        if entry.flags & Flags.SCALAR:
            array = entry.array.ravel()
        else:
            array = entry.array
        return array

        # [1] This is a workaround for old versions of numpy where calling masked_equal()
        #     on a string array fails with:
        #     numpy.core._exceptions.UFuncTypeError: ufunc 'equal' did not contain a loop
        #     with signature matching types (dtype('<U64'), dtype('<U64')) -> dtype('bool')

    def _get_subscript(self, entry, key):
        slice = self._get_slice(entry, key)
        slice_len = slice.stop - slice.start
        if key.rank:
            assert slice_len == 1
            subscript = slice.start
        elif slice_len == 1:
            max_level = self._node.max_levels[entry.name]
            if self._node.level == max_level:  # [1]
                subscript = slice.start
            else:
                subscript = slice
        else:
            subscript = slice
        return subscript

        # [1] Make sure we don't flatten single-rank array views unless they are
        #     accessed from the bottom-most DataBlock. This is important because
        #     otherwise shapes of array views could change depending on the values
        #     of replication factors. We want to provide 1d array views only in those
        #     cases where it's guaranteed there is always going to be only one rank for
        #     the given key in the current DataBlock, irrespective of replication
        #     factors.

    def _get_slice(self, entry, key):
        slice_ = self._slices[entry.name]
        if key.rank:
            rank_count = slice_.stop - slice_.start
            if key.rank > rank_count:
                message = (
                    "Rank %d is out of bounds; max. rank of '%s' in this view is %d"
                )
                raise NotFoundError(message % (key.rank, entry.name, rank_count))
            start = slice_.start + key.rank - 1
            slice_ = slice(start, start + 1)
        return slice_


def iterate_keys(node, index: MultiIndex) -> Iterator[Key]:
    """Recursively iterates over keys of `node` and all of its sub-nodes given
    the replication index.
    """
    if isinstance(node, LeafNode):
        yield from node.keys
    elif isinstance(node, ReplicationNode):
        if len(index) == node.factors.ndim + 1:
            for child in node.children:
                yield from iterate_keys(child, index)
        else:
            for i in range(node.factors[index]):
                child_index = index + (i,)
                for child in node.children:
                    yield from iterate_keys(child, child_index)
    elif isinstance(node, WrapperNode):
        for child in node.children:
            yield from iterate_keys(child, index)
    elif isinstance(node, AssociationNode):
        yield from node.keys
    else:
        assert False


def resolve_slices(node, index: MultiIndex) -> Dict[str, slice]:
    """Given the replication index, resolves ranks for each entry in the node."""
    try:
        slices = node.slices[index]
    except KeyError:
        slices = {}
        starts = resolve_starts(node, index)
        counts = resolve_counts(node, index)
        for name, count in counts.items():
            start = starts[name]
            stop = start + count
            slices[name] = slice(start, stop)
        node.slices[index] = slices
    return slices


def resolve_starts(node, index: MultiIndex) -> Counter:
    """Given the replication index, resolves rank starts for each entry in the node."""
    try:
        starts = node.starts[index]
    except KeyError:
        if not node.parent:
            starts = Counter()
        elif not isinstance(node, ReplicationNode):
            if node.ordinal > 0:
                upper_sibling = node.parent.children[node.ordinal - 1]
                starts = resolve_starts(upper_sibling, index).copy()
                counts = resolve_counts(upper_sibling, index)
                starts += counts
            else:
                starts = resolve_starts(node.parent, index)
        else:
            if index[-1] == 0:
                starts = resolve_starts(node.parent, index[:-1])
            else:
                lower_index = index[:-1] + (index[-1] - 1,)
                starts = resolve_starts(node, lower_index).copy()
                counts = resolve_counts(node, lower_index)
                starts += counts
        node.starts[index] = starts
    return starts


def resolve_counts(node, index: MultiIndex) -> Counter:
    """Given the replication index, resolves rank counts for each entry in the node."""
    try:
        counts = node.counts[index]
    except KeyError:
        if isinstance(node, (LeafNode, AssociationNode)):
            counts = Counter((k.name for k in node.keys))
            node.counts[index] = counts
        else:
            counts = Counter()
            for child in node.children:
                counts += resolve_counts(child, index)
            if len(index) == node.level:
                node.counts[index] = counts
    if (
        isinstance(node, LeafNode)
        and isinstance(node.parent, ReplicationNode)
        and len(index) < node.level
    ):
        total_factor = sum(flatten(node.parent.factors[index]))
        counts = counts.copy()
        for name in counts:
            counts[name] *= total_factor
    return counts


def resolve_elements(entries, root: Node):
    """Traverses nodes of the data tree and gathers lists of expanded elements
    for each of the entries.
    """

    def recurse(node, index):
        if isinstance(node, WrapperNode):
            for child in node.children:
                recurse(child, index)
        elif isinstance(node, ReplicationNode):
            assert node.level == len(index)
            for i in range(node.factors[index]):
                for child in node.children:
                    recurse(child, index + (i,))
        elif isinstance(node, (LeafNode, AssociationNode)):
            for key in node.keys:
                try:
                    entry = entries[key.name]
                except KeyError:  # [1]
                    continue
                if not entry.uniform_element:
                    entry.elements.append(key.element)
        else:
            assert False

    recurse(root, index=())

    # [1] Currently we append associated keys for each and every primary key
    #     of the LeafNode, but we add them to entries only if there is at least
    #     one rank covered by the bitmap. Check if this is OK or whether it needs
    #     to be optimized. TODO
